@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2021, University of Amsterdam
+    Copyright (c)  1985-2023, University of Amsterdam
                               VU University Amsterdam
                               SWI-Prolog Solutions b.v.
     All rights reserved.
@@ -88,20 +88,25 @@ version(Message) :-
                 *         INITIALISATION        *
                 *********************************/
 
-%!  load_init_file is det.
+%!  load_init_file(+ScriptMode) is det.
 %
 %   Load the user customization file. This can  be done using ``swipl -f
 %   file`` or simply using ``swipl``. In the   first  case we search the
 %   file both directly and over  the   alias  `user_app_config`.  In the
 %   latter case we only use the alias.
 
-load_init_file :-
+load_init_file(_) :-
     '$cmd_option_val'(init_file, OsFile),
     !,
     prolog_to_os_filename(File, OsFile),
     load_init_file(File, explicit).
-load_init_file :-
+load_init_file(prolog) :-
+    !,
     load_init_file('init.pl', implicit).
+load_init_file(none) :-
+    !,
+    load_init_file('init.pl', implicit).
+load_init_file(_).
 
 %!  loaded_init_file(?Base, ?AbsFile)
 %
@@ -225,6 +230,7 @@ prolog:message(initialize_now(Goal, Use)) -->
     ].
 
 '$run_initialization' :-
+    '$set_prolog_file_extension',
     '$run_initialization'(_, []),
     '$thread_init'.
 
@@ -343,14 +349,46 @@ path_sep -->
                  *   LOADING ASSIOCIATED FILES  *
                  *******************************/
 
-%!  argv_files(-Files) is det.
+%!  argv_prolog_files(-Files, -ScriptMode) is det.
 %
-%   Update the Prolog flag `argv`, extracting the leading script files.
+%   Update the Prolog flag `argv`, extracting  the leading script files.
+%   This is called after the C based  parser removed Prolog options such
+%   as ``-q``, ``-f none``, etc. These   options  are availabkle through
+%   '$cmd_option_val'/2.
+%
+%   Our task is to update the Prolog flag   `argv`  and return a list of
+%   the files to be loaded.   The rules are:
+%
+%     - If we find ``--`` all remaining options must go to `argv`
+%     - If we find *.pl files, these are added to Files and possibly
+%       remaining arguments are "script" arguments.
+%     - If we find an existing file, this is Files and possibly
+%       remaining arguments are "script" arguments.
+%     - File we find [search:]name, find search(name) as Prolog file,
+%       make this the content of `Files` and pass the remainder as
+%       options to `argv`.
+%
+%   @arg ScriptMode is one of
+%
+%     - exe
+%       Program is a saved state
+%     - prolog
+%       One or more *.pl files on commandline
+%     - script
+%       Single existing file on commandline
+%     - app
+%       [path:]cli-name on commandline
+%     - none
+%       Normal interactive session
 
-argv_files(Files) :-
+argv_prolog_files([], exe) :-
+    current_prolog_flag(saved_program_class, runtime),
+    !,
+    clean_argv.
+argv_prolog_files(Files, ScriptMode) :-
     current_prolog_flag(argv, Argv),
-    no_option_files(Argv, Argv1, Files, ScriptArgs),
-    (   (   ScriptArgs == true
+    no_option_files(Argv, Argv1, Files, ScriptMode),
+    (   (   nonvar(ScriptMode)
         ;   Argv1 == []
         )
     ->  (   Argv1 \== Argv
@@ -361,30 +399,55 @@ argv_files(Files) :-
         halt(1)
     ).
 
-no_option_files([--|Argv], Argv, [], true) :- !.
-no_option_files([Opt|_], _, _, ScriptArgs) :-
-    ScriptArgs \== true,
+no_option_files([--|Argv], Argv, [], ScriptMode) :-
+    !,
+    (   ScriptMode = none
+    ->  true
+    ;   true
+    ).
+no_option_files([Opt|_], _, _, ScriptMode) :-
+    var(ScriptMode),
     sub_atom(Opt, 0, _, _, '-'),
     !,
     '$usage',
     halt(1).
-no_option_files([OsFile|Argv0], Argv, [File|T], ScriptArgs) :-
+no_option_files([OsFile|Argv0], Argv, [File|T], ScriptMode) :-
     file_name_extension(_, Ext, OsFile),
     user:prolog_file_type(Ext, prolog),
     !,
-    ScriptArgs = true,
+    ScriptMode = prolog,
     prolog_to_os_filename(File, OsFile),
-    no_option_files(Argv0, Argv, T, ScriptArgs).
-no_option_files([OsScript|Argv], Argv, [Script], ScriptArgs) :-
-    ScriptArgs \== true,
+    no_option_files(Argv0, Argv, T, ScriptMode).
+no_option_files([OsScript|Argv], Argv, [Script], ScriptMode) :-
+    var(ScriptMode),
     !,
-    prolog_to_os_filename(Script, OsScript),
-    (   exists_file(Script)
+    prolog_to_os_filename(PlScript, OsScript),
+    (   exists_file(PlScript)
+    ->  Script = PlScript,
+        ScriptMode = script
+    ;   cli_script(OsScript, Script)
+    ->  ScriptMode = app,
+        set_prolog_flag(app_name, OsScript)
+    ;   '$existence_error'(file, PlScript)
+    ).
+no_option_files(Argv, Argv, [], ScriptMode) :-
+    (   ScriptMode = none
     ->  true
-    ;   '$existence_error'(file, Script)
+    ;   true
+    ).
+
+cli_script(CLI, Script) :-
+    (   sub_atom(CLI, Pre, _, Post, ':')
+    ->  sub_atom(CLI, 0, Pre, _, SearchPath),
+        sub_atom(CLI, _, Post, 0, Base),
+        Spec =.. [SearchPath, Base]
+    ;   Spec = app(CLI)
     ),
-    ScriptArgs = true.
-no_option_files(Argv, Argv, [], _).
+    absolute_file_name(Spec, Script,
+                       [ file_type(prolog),
+                         access(exist),
+                         file_errors(fail)
+                       ]).
 
 clean_argv :-
     (   current_prolog_flag(argv, [--|Argv])
@@ -392,20 +455,14 @@ clean_argv :-
     ;   true
     ).
 
-%!  associated_files(-Files)
+%!  win_associated_files(+Files)
 %
 %   If SWI-Prolog is started as <exe> <file>.<ext>, where <ext> is
 %   the extension registered for associated files, set the Prolog
 %   flag associated_file, switch to the directory holding the file
 %   and -if possible- adjust the window title.
 
-associated_files([]) :-
-    current_prolog_flag(saved_program_class, runtime),
-    !,
-    clean_argv.
-associated_files(Files) :-
-    '$set_prolog_file_extension',
-    argv_files(Files),
+win_associated_files(Files) :-
     (   Files = [File|_]
     ->  absolute_file_name(File, AbsFile),
         set_prolog_flag(associated_file, AbsFile),
@@ -419,7 +476,7 @@ associated_files(Files) :-
 %   When opening as a GUI application, e.g.,  by opening a file from
 %   the Finder/Explorer/..., we typically  want   to  change working
 %   directory to the location of  the   primary  file.  We currently
-%   detect that we are a GUI app  by the Prolog flag =console_menu=,
+%   detect that we are a GUI app  by the Prolog flag `console_menu`,
 %   which is set by swipl-win[.exe].
 
 set_working_directory(File) :-
@@ -444,8 +501,7 @@ set_window_title(_).
 
 %!  start_pldoc
 %
-%   If the option  =|--pldoc[=port]|=  is   given,  load  the  PlDoc
-%   system.
+%   If the option ``--pldoc[=port]`` is given, load the PlDoc system.
 
 start_pldoc :-
     '$cmd_option_val'(pldoc_server, Server),
@@ -510,28 +566,32 @@ initialise_error(E) :-
 
 initialise_prolog :-
     '$clean_history',
-    apple_setup_app,
+    apple_setup_app,                            % MacOS cwd/locale setup for swipl-win
     '$run_initialization',
-    '$load_system_init_file',
-    set_toplevel,
-    '$set_file_search_paths',
+    argv_prolog_files(Files, ScriptMode),
+    '$load_system_init_file',                   % -F file
+    set_toplevel,                               % set `toplevel_goal` flag from -t
+    '$set_file_search_paths',                   % handle -p alias=dir[:dir]*
     init_debug_flags,
-    start_pldoc,
+    start_pldoc,                                % handle --pldoc[=port]
     opt_attach_packs,
-    load_init_file,
+    load_init_file(ScriptMode),                 % -f file
     catch(setup_colors, E, print_message(warning, E)),
-    '$load_script_file',
-    associated_files(Files),
+    win_associated_files(Files),                % swipl-win: cd and update title
+    '$load_script_file',                        % -s file (may be repeated)
     load_associated_files(Files),
-    '$cmd_option_val'(goals, Goals),
-    (   Goals == [],
-        \+ '$init_goal'(when(_), _, _)
+    '$cmd_option_val'(goals, Goals),            % -g goal (may be repeated)
+    (   ScriptMode == app
+    ->  run_program_init,                       % initialization(Goal, program)
+        run_main_init(true)
+    ;   Goals == [],
+        \+ '$init_goal'(when(_), _, _)          % no -g or -t or initialization(program)
     ->  version                                 % default interactive run
-    ;   run_init_goals(Goals),
-        (   load_only
+    ;   run_init_goals(Goals),                  % run -g goals
+        (   load_only                           % used -l to load
         ->  version
-        ;   run_program_init,
-            run_main_init
+        ;   run_program_init,                   % initialization(Goal, program)
+            run_main_init(false)                % initialization(Goal, main)
         )
     ).
 
@@ -605,7 +665,7 @@ run_program_init :-
     forall('$init_goal'(when(program), Goal, Ctx),
            run_init_goal(Goal, @(Goal,Ctx))).
 
-run_main_init :-
+run_main_init(_) :-
     findall(Goal-Ctx, '$init_goal'(when(main), Goal, Ctx), Pairs),
     '$last'(Pairs, Goal-Ctx),
     !,
@@ -614,7 +674,9 @@ run_main_init :-
     ;   true
     ),
     run_init_goal(Goal, @(Goal,Ctx)).
-run_main_init.
+run_main_init(true) :-
+    '$existence_error'(initialization, main).
+run_main_init(_).
 
 run_init_goal(Goal, Ctx) :-
     (   catch_with_backtrace(user:Goal, E, true)
@@ -936,16 +998,20 @@ read_query(Prompt, Goal, Bindings) :-
 
 %!  read_query_line(+Input, -Line) is det.
 
+read_query_line(Input, Line) :-
+    stream_property(Input, error(true)),
+    !,
+    Line = end_of_file.
 read_query_line(Input, Line) :-
     catch(read_term_as_atom(Input, Line), Error, true),
     save_debug_after_read,
     (   var(Error)
     ->  true
-    ;   Error = error(syntax_error(_),_)
-    ->  print_message(error, Error),
-        fail
-    ;   print_message(error, Error),
-        throw(Error)
+    ;   catch(print_message(error, Error), _, true),
+        (   Error = error(syntax_error(_),_)
+        ->  fail
+        ;   throw(Error)
+        )
     ).
 
 %!  read_term_as_atom(+Input, -Line)
