@@ -3,9 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org/projects/xpce/
-    Copyright (c)  2006-2020, University of Amsterdam
+    Copyright (c)  2006-2023, University of Amsterdam
                               VU University Amsterdam
                               CWI, Amsterdam
+                              SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -69,7 +70,7 @@
             xref_defined_class/3        % ?Source, ?ClassName, -How
           ]).
 :- autoload(library(apply),[maplist/2,partition/4,maplist/3]).
-:- autoload(library(debug),[debug/3]).
+:- use_module(library(debug),[debug/3]).
 :- autoload(library(dialect),[expects_dialect/1]).
 :- autoload(library(error),[must_be/2,instantiation_error/1]).
 :- autoload(library(lists),[member/2,append/2,append/3,select/3]).
@@ -77,12 +78,14 @@
 :- autoload(library(operators),[push_op/3]).
 :- autoload(library(option),[option/2,option/3]).
 :- autoload(library(ordsets),[ord_intersect/2,ord_intersection/3]).
+:- autoload(library(prolog_code), [pi_head/2]).
 :- autoload(library(prolog_source),
 	    [ prolog_canonical_source/2,
 	      prolog_open_source/2,
 	      prolog_close_source/1,
 	      prolog_read_source_term/4
 	    ]).
+
 :- if(exists_source(library(shlib))).
 :- autoload(library(shlib),[current_foreign_library/2]).
 :- endif.
@@ -91,6 +94,7 @@
 :- if(exists_source(library(pldoc))).
 :- use_module(library(pldoc), []).      % Must be loaded before doc_process
 :- use_module(library(pldoc/doc_process)).
+
 :- endif.
 
 :- predicate_options(xref_source/2, 2,
@@ -123,7 +127,7 @@
     (mode)/2,                       % Mode, Src
     xoption/2,                      % Src, Option
     xflag/4,                        % Name, Value, Src, Line
-
+    grammar_rule/2,                 % Head, Src
     module_comment/3,               % Src, Title, Comment
     pred_comment/4,                 % Head, Src, Summary, Comment
     pred_comment_link/3,            % Head, Src, HeadTo
@@ -421,6 +425,20 @@ set_xref(Xref) :-
     current_prolog_flag(xref, Xref),
     set_prolog_flag(xref, true).
 
+:- meta_predicate
+    with_xref(0).
+
+with_xref(Goal) :-
+    current_prolog_flag(xref, Xref),
+    (   Xref == true
+    ->  call(Goal)
+    ;   setup_call_cleanup(
+            set_prolog_flag(xref, true),
+            Goal,
+            set_prolog_flag(xref, Xref))
+    ).
+
+
 %!  set_initial_mode(+Stream, +Options) is det.
 %
 %   Set  the  initial  mode  for  processing    this   file  in  the
@@ -526,6 +544,7 @@ xref_clean(Source) :-
     retractall(uses_file(_, Src, _)),
     retractall(xmodule(_, Src)),
     retractall(xop(Src, _)),
+    retractall(grammar_rule(_, Src)),
     retractall(xoption(Src, _)),
     retractall(xflag(_Name, _Value, Src, Line)),
     retractall(source(Src, _)),
@@ -607,6 +626,7 @@ xref_called(Source, Called, By, Cond, Line) :-
 %     * foreign(Location)
 %     * constraint(Location)
 %     * imported(From)
+%     * dcg
 
 xref_defined(Source, Called, How) :-
     nonvar(Source),
@@ -633,6 +653,8 @@ xref_defined2(constraint(Line), Src, Called) :-
     constraint(Called, Src, Line).
 xref_defined2(imported(From), Src, Called) :-
     imported(Called, Src, From).
+xref_defined2(dcg, Src, Called) :-
+    grammar_rule(Called, Src).
 
 
 %!  xref_definition_line(+How, -Line)
@@ -722,7 +744,8 @@ xref_defined_class(Source, Class, file(File)) :-
 
 :- thread_local
     current_cond/1,
-    source_line/1.
+    source_line/1,
+    current_test_unit/2.
 
 current_source_line(Line) :-
     source_line(Var),
@@ -746,7 +769,8 @@ collect(Src, File, In, Options) :-
         Comments = []
     ;   CommentHandling == store
     ->  CommentOptions = [ process_comment(true) ],
-        Comments = []
+        Comments = [],
+	set_prolog_flag(xref_store_comments, true)
     ;   CommentOptions = [ comments(Comments) ]
     ),
     repeat,
@@ -764,7 +788,8 @@ collect(Src, File, In, Options) :-
                   E, print_message(error, E)),
             erase(Ref)),
         EOF == true,
-    !.
+    !,
+    set_prolog_flag(xref_store_comments, false).
 
 report_syntax_error(E, _, _) :-
     fatal_error(E),
@@ -862,6 +887,13 @@ process(Term, Comments, Term0, TermPos, Src, false) :-
 process(_, Term0, _) :-
     ignore_raw_term(Term0),
     !.
+process(Head :- Body, Head0 --> _, Src) :-
+    pi_head(F/A, Head),
+    pi_head(F/A0, Head0),
+    A =:= A0 + 2,
+    !,
+    assert_grammar_rule(Src, Head),
+    process((Head :- Body), Src).
 process(Term, _Term0, Src) :-
     process(Term, Src).
 
@@ -1063,6 +1095,12 @@ process_directive(module(Module, Export, Import), Src) :-
     assert_module(Src, Module),
     assert_module_export(Src, Export),
     assert_module3(Import, Src).
+process_directive(begin_tests(Unit, _Options), Src) :-
+    enter_test_unit(Unit, Src).
+process_directive(begin_tests(Unit), Src) :-
+    enter_test_unit(Unit, Src).
+process_directive(end_tests(Unit), Src) :-
+    leave_test_unit(Unit, Src).
 process_directive('$set_source_module'(system), Src) :-
     assert_module(Src, system).     % hack for handling boot/init.pl
 process_directive(pce_begin_class_definition(Name, Meta, Super, Doc), Src) :-
@@ -1219,7 +1257,7 @@ xref_meta_src(Head, Called, _) :-
     arg(1, Head, G),
     Called = [G+Extra].
 xref_meta_src(Head, Called, _) :-
-    predicate_property(user:Head, meta_predicate(Meta)),
+    with_xref(predicate_property('$xref_tmp':Head, meta_predicate(Meta))),
     !,
     Meta =.. [_|Args],
     meta_args(Args, 1, Head, Called).
@@ -1440,7 +1478,7 @@ hook(user:prolog_clause_name(_,_)).
 hook(user:prolog_list_goal(_)).
 hook(user:prolog_predicate_name(_,_)).
 hook(user:prolog_trace_interception(_,_,_,_)).
-hook(user:prolog_exception_hook(_,_,_,_)).
+hook(prolog:prolog_exception_hook(_,_,_,_,_)).
 hook(sandbox:safe_primitive(_)).
 hook(sandbox:safe_meta_predicate(_)).
 hook(sandbox:safe_meta(_,_)).
@@ -1482,6 +1520,9 @@ process_body(Body, Origin, Src) :-
 process_goal(Var, _, _, _) :-
     var(Var),
     !.
+process_goal(_:Goal, _, _, _) :-
+    var(Goal),
+    !.
 process_goal(Goal, Origin, Src, P) :-
     Goal = (_,_),                               % problems
     !,
@@ -1498,7 +1539,12 @@ process_goal(Goal, Origin, Src, P) :-
         ->  true
         ;   M = user
         ),
-        (   predicate_property(M:Goal, imported_from(IM))
+        pi_head(PI, M:Goal),
+        (   current_predicate(PI),
+            predicate_property(M:Goal, imported_from(IM))
+        ->  true
+        ;   PI = M:Name/Arity,
+            '$find_library'(M, Name, Arity, IM, _Library)
         ->  true
         ;   IM = M
         ),
@@ -1720,6 +1766,17 @@ partial_evaluate(_, _).
 eval(X = Y) :-
     unify_with_occurs_check(X, Y).
 
+		 /*******************************
+		 *        PLUNIT SUPPORT	*
+		 *******************************/
+
+enter_test_unit(Unit, _Src) :-
+    current_source_line(Line),
+    asserta(current_test_unit(Unit, Line)).
+
+leave_test_unit(Unit, _Src) :-
+    retractall(current_test_unit(Unit, _)).
+
 
                  /*******************************
                  *          XPCE STUFF          *
@@ -1901,6 +1958,8 @@ prolog:no_autoload_module(library(pldoc)).
 prolog:no_autoload_module(library(settings)).
 prolog:no_autoload_module(library(debug)).
 prolog:no_autoload_module(library(plunit)).
+prolog:no_autoload_module(library(macros)).
+prolog:no_autoload_module(library(yall)).
 
 
 %!  process_requires(+Import, +Src)
@@ -2470,6 +2529,16 @@ expand_hide_called(pce_principal:get_implementation(_, _, _, _)).
 expand_hide_called(pce_principal:pce_lazy_get_method(_,_,_)).
 expand_hide_called(pce_principal:pce_lazy_send_method(_,_,_)).
 
+assert_defined(Src, Goal) :-
+    Goal = test(_Test),
+    current_test_unit(Unit, Line),
+    assert_called(Src, '<test_unit>'(Unit), Goal, Line),
+    fail.
+assert_defined(Src, Goal) :-
+    Goal = test(_Test, _Options),
+    current_test_unit(Unit, Line),
+    assert_called(Src, '<test_unit>'(Unit), Goal, Line),
+    fail.
 assert_defined(Src, Goal) :-
     defined(Goal, Src, _),
     !.
@@ -2486,6 +2555,14 @@ assert_foreign(Src, Goal) :-
     current_source_line(Line),
     assert(foreign(Term, Src, Line)).
 
+assert_grammar_rule(Src, Goal) :-
+    grammar_rule(Goal, Src),
+    !.
+assert_grammar_rule(Src, Goal) :-
+    generalise(Goal, Term),
+    assert(grammar_rule(Term, Src)).
+
+
 %!  assert_import(+Src, +Import, +ExportList, +From, +Reexport) is det.
 %
 %   Asserts imports into Src. Import   is  the import specification,
@@ -2769,6 +2846,15 @@ generalise(pce_principal:get_implementation(Id, _, _, _),
     atom(Id),
     !.
 generalise('<directive>'(Line), '<directive>'(Line)) :- !.
+generalise(test(Test), test(Test)) :-
+    current_test_unit(_,_),
+    ground(Test),
+    !.
+generalise(test(Test, _), test(Test, _)) :-
+    current_test_unit(_,_),
+    ground(Test),
+    !.
+generalise('<test_unit>'(Line), '<test_unit>'(Line)) :- !.
 generalise(Module:Goal0, Module:Goal) :-
     atom(Module),
     !,
