@@ -60,9 +60,8 @@ how the various predicates can be called from Prolog.
 #include <iostream>
 #include <sstream>
 #include <memory>
-#include <SWI-Stream.h>
-#include <SWI-Prolog.h>
 #include "SWI-cpp2.h"
+#include "SWI-cpp2-atommap.h"
 #include <errno.h>
 #include <math.h>
 #include <cassert>
@@ -83,6 +82,17 @@ using namespace std;
 #endif
 
 
+PREDICATE(unwrap, 1)
+{ // test that the definition of C_ as a macro hasn't broken anything
+  PlTerm A1_copy(A1);
+  PlTerm *A1_ptr = &A1_copy;
+
+  return A1.C_ == A1.unwrap() &&
+    A1.C_ == *PlUnwrapAsPtr(A1_ptr) &&
+    A1.unwrap() == *PlUnwrapAsPtr(A1_ptr);
+}
+
+
 PREDICATE(hello, 0)
 { PlQuery q("write", PlTermv(PlTerm_atom("hello hello hello")));
   PlCheckFail(q.next_solution());
@@ -94,9 +104,9 @@ PREDICATE(hello, 2)
   // This will result in an encoding error if A1 isn't Latin-1
   buffer << "Hello " << A1.as_string() << endl;
   buffer << "Hello " << A1.as_string().c_str() << endl; // Same output as previous line
-  buffer << "Hello " << A1.as_string(PlEncoding::Latin1).c_str() << endl; // Also same, if it's ASCII
-  buffer << "Hello " << A1.as_string(PlEncoding::UTF8).c_str() << endl;
-  buffer << "Hello " << A1.as_string(PlEncoding::Locale).c_str() << endl; // Can vary by locale settings
+  buffer << "Hello " << A1.as_string(PlEncoding::Latin1) << endl; // Also same, if it's ASCII
+  buffer << "Hello " << A1.as_string(PlEncoding::UTF8) << endl;
+  buffer << "Hello " << A1.as_string(PlEncoding::Locale) << endl; // Can vary by locale settings
 
   return A2.unify_string(buffer.str());
 }
@@ -107,9 +117,9 @@ PREDICATE(hello2, 2)
   // The following have the same output as hello/1, if A1 is an atom
   buffer << "Hello2 " << atom_a1.as_string() << endl;
   buffer << "Hello2 " << A1.as_string().c_str() << endl;
-  buffer << "Hello2 " << A1.as_string(PlEncoding::Latin1).c_str() << endl;
-  buffer << "Hello2 " << A1.as_string(PlEncoding::UTF8).c_str() << endl;
-  buffer << "Hello2 " << A1.as_string(PlEncoding::Locale).c_str() << endl;
+  buffer << "Hello2 " << A1.as_string(PlEncoding::Latin1) << endl;
+  buffer << "Hello2 " << A1.as_string(PlEncoding::UTF8) << endl;
+  buffer << "Hello2 " << A1.as_string(PlEncoding::Locale) << endl;
 
   return A2.unify_string(buffer.str());
 }
@@ -140,11 +150,17 @@ PREDICATE(add, 3)
 
 PREDICATE(add_num, 3)
 { auto x = A1, y = A2, result = A3;
-  // Note that as_float() handles floats
+  // Note that as_float() handles integers
   double sum = x.as_float() + y.as_float();
-  if ( double(long(sum)) == sum ) /* Can float be represented as int? */
-    return result.unify_integer(long(sum));
-  return result.unify_float(sum);
+  return ( double(long(sum)) == sum ) // Can float be represented as int?
+    ? result.unify_integer(long(sum))
+    : result.unify_float(sum);
+}
+
+PREDICATE(name_arity, 1)
+{ PlStream strm(Scurrent_output);
+  strm.printf("name = %s, arity = %zd\n", A1.name().as_string().c_str(), A1.arity());
+  return true;
 }
 
 PREDICATE(name_arity, 2)
@@ -165,6 +181,17 @@ PREDICATE(name_arity, 3)		/* name_arity(+Term, -Name, -Arity) */
   return true;
 }
 
+PREDICATE(name_arity_bool, 3)  // like name_arity/3 but doesn't throw
+{ auto term = A1; auto name = A2; auto arity = A3;
+  PlAtom name_a(PlAtom::null);
+  size_t arity_a;
+  if ( !term.name_arity(&name_a, &arity_a) )
+    return false;
+  assert(name_a.not_null());
+
+  return name.unify_atom(name_a) && arity.unify_integer(arity_a);
+}
+
 PREDICATE(list_modules, 1)
 { std::stringstream buffer;
   PlTermv av(1);
@@ -177,6 +204,8 @@ PREDICATE(list_modules, 1)
   return A1.unify_string(buffer.str());
 }
 
+// %! average(+Templ, :Goal, -Average) is det.
+// % Same as: aggregate(sum(X)/count, Goal, A), Average is A.
 PREDICATE(average, 3)			/* average(+Templ, :Goal, -Average) */
 { long sum = 0;
   long n = 0;
@@ -246,7 +275,7 @@ PREDICATE(term, 2)
 { static PlAtom ATOM_atom("atom");
   PlAtom a(A1.as_atom());
 
-  if ( a.C_ == ATOM_atom.C_ )
+  if ( a.unwrap() == ATOM_atom.unwrap() )
     return A2.unify_atom("hello world"); // or A2.unify_term(PlAtom("hello world"));
   if ( A1.as_string() == "string" )
     return A2.unify_string("hello world");
@@ -279,17 +308,24 @@ PREDICATE(eq2, 2)
 }
 
 PREDICATE(eq3, 2)
-{ PlCheckFail(PL_unify(A1.C_, A2.C_));
-  return true;
+{ return Plx_unify(A1.unwrap(), A2.unwrap());
+}
+
+PREDICATE(eq4, 2)
+{ // This is what Plx_unify() expands to
+  return PlWrap<bool>(PL_unify(A1.unwrap(), A2.unwrap()));
 }
 
 PREDICATE(write_list, 1)
 { PlTerm_tail tail(A1);
   PlTerm_var e;
-
-  while(tail.next(e))
-    cout << e.as_string() << endl;
-
+  PlStream strm(Scurrent_output);
+  // This is an example of using the try...PREDICATE_CATCH for avoiding
+  // throwing an exception in the PlStream destructor.
+  try
+  { while( tail.next(e) )
+      strm.printf("%s\n", e.as_string().c_str());
+  } PREDICATE_CATCH({strm.release(); return false;})
   return true;
 }
 
@@ -298,7 +334,7 @@ PREDICATE(cappend, 3)
   PlTerm_tail l3(A3);
   PlTerm_var e;
 
-  while(l1.next(e))
+  while( l1.next(e) )
     PlCheckFail(l3.append(e));
 
   return A2.unify_term(l3);
@@ -313,6 +349,7 @@ PREDICATE(cpp_call_, 3)
 { int flags = A2.as_int();
   int verbose = A3.as_bool();
   std::string flag_str;
+  PlStream strm(Scurrent_output);
   // if ( flags & PL_Q_DEBUG )        flag_str.append(",debug");
   // if ( flags & PL_Q_DETERMINISTIC) flag_str.append(",deterministic");
   if ( flags & PL_Q_NORMAL )          flag_str.append(",normal");
@@ -326,7 +363,7 @@ PREDICATE(cpp_call_, 3)
   else
     flag_str = std::string("cpp_call(").append(flag_str.substr(1)).append(")");
   if ( verbose )
-    cout << flag_str << ": " << A1.as_string() << endl;
+    strm.printf("%s: %s\n",  flag_str.c_str(), A1.as_string().c_str());
 
   try
   { int rc = PlCall(A1, flags);
@@ -341,33 +378,33 @@ PREDICATE(cpp_call_, 3)
 	default:             status_str = "???";       break;
       }
       if ( verbose )
-	cout << "... after call, rc=" << rc << ": " << status_str << endl;
+	strm.printf("... after call, rc=%d: %s\n", rc, status_str);
     } else
     { if ( verbose )
-	cout << "... after call, rc=" << rc << endl;
+	strm.printf("... after call, rc=%d\n", rc);
     }
 
     if ( rc )
     { if ( verbose )
-	cout << "cpp_call result: rc=" << rc << ": " << A1.as_string() << endl;
+	strm.printf("cpp_call result: rc=%d: %s\n", rc, A1.as_string().c_str());
     } else
-    { PlTerm_term_t ex(Plx_exception(0));
+    { PlTerm ex(Plx_exception(0));
       if ( ex.is_null() )
       { if ( verbose )
-	  cout << "cpp_call failed" << endl;
+	  strm.printf("cpp_call failed\n");
       } else
       { if ( verbose )
-	  cout << "cpp_call failed: ex: " << ex.as_string() << endl;
+	  strm.printf("cpp_call failed: ex: %s\n", ex.as_string().c_str());
       }
     }
     return rc; // TODO: this is wrong with some query flags
   } catch ( PlException& ex )
   { if ( ex.is_null() )
     { if ( verbose )
-	cout << "cpp_call except is_null" << endl;
+	strm.printf("cpp_call except is_null\n");
     } else
     { if ( verbose )
-	cout << "cpp_call exception: " << ex.as_string() << endl;
+	strm.printf("cpp_call exception: %s\n", ex.as_string().c_str());
     }
     throw;
   }
@@ -375,12 +412,13 @@ PREDICATE(cpp_call_, 3)
 
 PREDICATE(cpp_atom_codes, 2)
 { int rc = PlCall("atom_codes", PlTermv(A1, A2));
-  if ( ! rc )
-  { PlException ex(PlTerm_term_t(Plx_exception(0)));
+  if ( !rc )
+  { PlException ex(PlTerm(Plx_exception(0)));
+    PlStream strm(Scurrent_output);
     if ( ex.is_null() )
-      cout << "atom_codes failed" << endl;
+      strm.printf("atom_codes failed\n");
     else
-      cout << "atom_codes failed: ex: " << ex.as_string() << endl; // Shouldn't happen
+      strm.printf("atom_codes failed: ex: %s\n", ex.as_string().c_str()); // Shouldn't happen
   }
   return rc;
 }
@@ -400,7 +438,7 @@ PREDICATE(square_roots, 2)
 { int end = A1.as_int();
   PlTerm_tail list(A2);
 
-  for(int i=0; i<end; i++)
+  for(int i=0; i<=end; i++)
     PlCheckFail(list.append(PlTerm_float(sqrt(double(i)))));
 
   return list.close();
@@ -476,7 +514,7 @@ PREDICATE(free_my_object, 1)
 }
 
 PREDICATE(make_functor, 3)  // make_functor(foo, x, foo(x))
-{ auto f = PlFunctor(A1.as_atom().as_string().c_str(), 1);
+{ auto f = PlFunctor(A1.as_atom().as_string(), 1);
   return A3.unify_functor(f) &&
     A3[1].unify_term(A2);
 }
@@ -566,6 +604,8 @@ PREDICATE(ensure_PlTerm_forward_declarations_are_implemented, 0)
   PlTerm_atom p_atom5(std::wstring(L"世界"));
   PlTerm_term_t t_t(Plx_new_term_ref());
   PlTerm_term_t t_null(PlTerm::null);
+  PlTerm t_t2(Plx_new_term_ref());
+  PlTerm t_null2(PlTerm::null);
   // The various integer types are also used in IntInfo.
   PlTerm_integer t_int1(std::numeric_limits<int>::max());
   PlTerm_integer t_int1b(std::numeric_limits<int>::min());
@@ -743,29 +783,29 @@ PREDICATE(unify_int_set, 1)
 
 // The following are for verifying some documentation details.
 
-PREDICATE(c_PL_unify_nil, 1)          { return static_cast<foreign_t>(PL_unify_nil(A1.C_)); }
+PREDICATE(c_PL_unify_nil, 1)          { return static_cast<foreign_t>(PL_unify_nil(A1.unwrap())); }
 
 PREDICATE(cpp_unify_nil, 1)           { return A1.unify_nil(); }
 
-PREDICATE(check_c_PL_unify_nil, 1)    { PlEx<bool>(PL_unify_nil(A1.C_)); return true; }
+PREDICATE(check_c_PL_unify_nil, 1)    { PlEx<bool>(PL_unify_nil(A1.unwrap())); return true; }
 
 // Repeat the above, for *_ex():
 
-PREDICATE(c_PL_unify_nil_ex, 1)       { return static_cast<foreign_t>(PL_unify_nil_ex(A1.C_)); }
+PREDICATE(c_PL_unify_nil_ex, 1)       { return static_cast<foreign_t>(PL_unify_nil_ex(A1.unwrap())); }
 
 PREDICATE(cpp_unify_nil_ex, 1)        { A1.unify_nil_ex(); return true; }
 
-PREDICATE(check_c_PL_unify_nil_ex, 1) { PlEx<bool>(PL_unify_nil_ex(A1.C_)); return true; }
+PREDICATE(check_c_PL_unify_nil_ex, 1) { PlEx<bool>(PL_unify_nil_ex(A1.unwrap())); return true; }
 
 
 
-PREDICATE(c_PL_get_nil, 1)            { return static_cast<foreign_t>(PL_get_nil(A1.C_)); }
+PREDICATE(c_PL_get_nil, 1)            { return static_cast<foreign_t>(PL_get_nil(A1.unwrap())); }
 
 PREDICATE(cpp_as_nil, 1)              { A1.as_nil();                     return true; }
 
-PREDICATE(check_c_PL_get_nil, 1)      { PlEx<bool>(PL_get_nil(A1.C_));      return true; }
+PREDICATE(check_c_PL_get_nil, 1)      { PlEx<bool>(PL_get_nil(A1.unwrap()));      return true; }
 
-PREDICATE(check_c_PL_get_nil_ex, 1)   { PlEx<bool>(PL_get_nil_ex(A1.C_));   return true; }
+PREDICATE(check_c_PL_get_nil_ex, 1)   { PlEx<bool>(PL_get_nil_ex(A1.unwrap()));   return true; }
 
 // Functions re-implemented from ffi4pl.c
 
@@ -811,15 +851,16 @@ PREDICATE_NONDET(range_cpp, 3)
 }
 
 
-
 // For benchmarking `throw PlThrow()` vs `return false`
 // Times are given for 1 million failures
 
-// 0.085 sec for ime((between(1,1000000,X), fail)).
-
+// 0.085 sec for time((between(1,1000000,X), fail)).
 // 0.16 sec for time((between(1,1000000,X), X=0)).
-
 // 0.20 sec for time((between(1,1000000,X), unify_zero_0(X))).
+
+// See also timings for individual unify_zero_* predicates, unify_foo_*
+
+
 static foreign_t
 unify_zero_0(term_t a1)
 { return static_cast<foreign_t>(Plx_unify_integer(a1, 0));
@@ -835,21 +876,21 @@ static PlRegister _x_unify_zero_4_1(nullptr, "unify_zero_0", unify_zero_0);
 
 // 0.23 sec for time((between(1,1000000,X), unify_zero_1(X))).
 PREDICATE(unify_zero_1, 1)
-{ if ( !Plx_unify_integer(A1.C_, 0) )
+{ if ( !Plx_unify_integer(A1.unwrap(), 0) )
     return false;
   return true;
 }
 
 // 3.3 sec for time((between(1,1000000,X), unify_zero_2(X))).
 PREDICATE(unify_zero_2, 1)
-{ if ( !Plx_unify_integer(A1.C_, 0) )
+{ if ( !Plx_unify_integer(A1.unwrap(), 0) )
     throw PlFail();
   return true;
 }
 
 // 4.0 sec for time((between(1,1000000,X), unify_zero_3(X))).
 PREDICATE(unify_zero_3, 1)
-{ PlCheckFail( Plx_unify_integer(A1.C_, 0) );
+{ PlCheckFail( Plx_unify_integer(A1.unwrap(), 0) );
   return true;
 }
 
@@ -913,6 +954,40 @@ PREDICATE(unify_foo_string_2b, 1)
 
 // end of benchmarking predicates
 
+// Example pl_write_atoms from foreign.doc
+
+PREDICATE(pl_write_atoms_cpp, 1)
+{ auto l = A1;
+  PlTerm_var head;
+  PlTerm tail(l.copy_term_ref());
+  PlStream strm(Scurrent_output);
+
+  while( tail.get_list_ex(head, tail) )
+  { strm.printf("%s\n", head.as_string().c_str());
+  }
+
+  tail.get_nil_ex();
+  return true;
+}
+
+PREDICATE(pl_write_atoms_c, 1)
+{ term_t l = A1.unwrap();
+  term_t head = PL_new_term_ref();   /* the elements */
+  term_t tail = PL_copy_term_ref(l); /* copy (we modify tail) */
+  int rc = TRUE;
+  auto strm = Plx_acquire_stream(Scurrent_output);
+
+  while( rc && PL_get_list_ex(tail, head, tail) )
+  { PL_STRINGS_MARK();
+      char *s;
+      if ( (rc=PL_get_chars(head, &s, CVT_ATOM|REP_MB|CVT_EXCEPTION)) )
+        Sfprintf(strm, "%s\n", s);
+    PL_STRINGS_RELEASE();
+  }
+  Plx_release_stream(strm);
+
+  return rc && PL_get_nil_ex(tail); /* test end for [] */
+}
 
 // Predicates for checking native integer handling
 // See https://en.cppreference.com/w/cpp/types/numeric_limits
@@ -1051,14 +1126,13 @@ PREDICATE(type_error_string, 3)
 PREDICATE(w_atom_cpp_, 2)
 { auto stream = A1, t = A2;
   IOSTREAM* s;
-  Plx_get_stream(stream.C_, &s, SIO_INPUT);
-  { PlStringBuffers _string_buffers;
-    size_t len;
-    const pl_wchar_t *sa = Plx_atom_wchars(t.as_atom().C_, &len);
-    // TODO: Sfprintf() doesn't get format checked in C++
-    Sfprintf(s, "/%Ws/%zd", sa, len);
-  }
-  return TRUE;
+  Plx_get_stream(stream.unwrap(), &s, SIO_INPUT);
+  PlStream strm(Scurrent_output);
+  PlStringBuffers _string_buffers;
+  size_t len;
+  const pl_wchar_t *sa = Plx_atom_wchars(t.as_atom().unwrap(), &len);
+  strm.printfX("/%Ws/%zd", sa, len);
+  return true;
 }
 
 
@@ -1089,8 +1163,8 @@ PREDICATE(cpp_options, 3)
   };
 
   PlStringBuffers _string_buffers; // for descr's contents
-  PlEx<bool>(PL_scan_options(options.C_, flags, "cpp_options", scan_options,
-                             &quoted, &length, &callback.C_, &token.C_, &descr));
+  PlEx<bool>(PL_scan_options(options.unwrap(), flags, "cpp_options", scan_options,
+                             &quoted, &length, callback.unwrap_as_ptr(), token.unwrap_as_ptr(), &descr));
 
   PlCheckFail(result.unify_term(
                   PlCompound("options",
@@ -1115,7 +1189,7 @@ PREDICATE(cvt_i_bool, 2)
 // TODO: add PlEngine tests
 
 PREDICATE(throw_domain_cpp0, 1)
-{ return Plx_domain_error("footype", A1.C_);
+{ return Plx_domain_error("footype", A1.unwrap());
 }
 
 PREDICATE(throw_domain_cpp1, 1)
@@ -1123,12 +1197,12 @@ PREDICATE(throw_domain_cpp1, 1)
 }
 
 PREDICATE(throw_domain_cpp2, 1)
-{ PlEx<bool>(Plx_domain_error("footype", A1.C_));
+{ PlEx<bool>(Plx_domain_error("footype", A1.unwrap()));
   return false; // Should never reach here
 }
 
 PREDICATE(throw_domain_cpp3, 1)
-{ if ( !Plx_domain_error("footype", A1.C_) )
+{ if ( !Plx_domain_error("footype", A1.unwrap()) )
     throw PlFail();
   // Shouldn't fall through to here
   Plx_clear_exception();
@@ -1148,27 +1222,45 @@ PREDICATE(throw_uninstantiation_error_cpp, 1)
 }
 
 PREDICATE(throw_representation_error_cpp, 1)
-{ throw PlRepresentationError(A1.as_string().c_str());
+{ throw PlRepresentationError(A1.as_string());
 }
 
 PREDICATE(throw_type_error_cpp, 2)
-{ throw PlTypeError(A1.as_string().c_str(), A2);
+{ throw PlTypeError(A1.as_string(), A2);
+}
+
+PREDICATE(throw_and_check_error_cpp, 2)
+{ try
+  { throw PlTypeError(A1.as_string(), A2);
+  } catch (const PlException& e)
+  { PlAtom ATOM_error("error");
+    PlAtom ATOM_type_error("type_error");
+    PlTerm e_t = e.term();
+    // if A1 is 'atom', then e_t is error(type_error(atom,A2),_):
+    PlCheckFail(e_t.name() == ATOM_error);
+    PlCheckFail(e_t.arity() == 2);
+    PlCheckFail(e_t[1].name() == ATOM_type_error);
+    PlCheckFail(e_t[1][1].as_string() == A1.as_string());
+    PlCheckFail(e_t[1][2] == A2);
+    throw;
+  }
+  return true;
 }
 
 PREDICATE(throw_domain_error_cpp, 2)
-{ throw PlDomainError(A1.as_string().c_str(), A2);
+{ throw PlDomainError(A1.as_string(), A2);
 }
 
 PREDICATE(throw_existence_error_cpp, 2)
-{ throw PlExistenceError(A1.as_string().c_str(), A2);
+{ throw PlExistenceError(A1.as_string(), A2);
 }
 
 PREDICATE(throw_permission_error_cpp, 3)
-{ throw PlPermissionError(A1.as_string().c_str(), A2.as_string().c_str(), A3);
+{ throw PlPermissionError(A1.as_string(), A2.as_string(), A3);
 }
 
 PREDICATE(throw_resource_error_cpp, 1)
-{ throw PlResourceError(A1.as_string().c_str());
+{ throw PlResourceError(A1.as_string());
 }
 
 PREDICATE(ten, 10)
@@ -1265,10 +1357,11 @@ struct MyBlob : public PlBlob
   }
 
   bool write_fields(IOSTREAM *s, int flags) const override
-  { if ( Sfprintf(s, ",name=%s", name_.c_str()) < 0 )
-      return false;
+  { PlStream strm(s);
+
+    strm.printf(",name=%s", name_.c_str());
     if ( !connection )
-      return Sfprintf(s, ",closed") < 0;
+      strm.printf(",closed");
     return true;
   }
 };
@@ -1289,3 +1382,49 @@ PREDICATE(close_my_blob, 1)
     throw ref->MyBlobError("my_blob_close_error");
   return true;
 }
+
+
+static AtomMap<PlAtom, PlAtom> map_atom_atom("add", "atom_atom");
+
+PREDICATE(atom_atom_find, 2)
+{ auto value = map_atom_atom.find(A1.as_atom());
+  PlCheckFail(value.not_null());
+  return A2.unify_atom(value);
+}
+
+PREDICATE(atom_atom_add, 2)
+{ map_atom_atom.insert(A1.as_atom(), A2.as_atom());
+  return true;
+}
+
+PREDICATE(atom_atom_erase, 1)
+{ map_atom_atom.erase(A1.as_atom());
+  return true;
+}
+
+PREDICATE(atom_atom_size, 1)
+{ return A1.unify_integer(map_atom_atom.size());
+}
+
+static AtomMap<PlTerm, PlRecord> map_atom_term("insert", "atom_term");
+
+PREDICATE(atom_term_find, 2)
+{ auto value = map_atom_term.find(A1.as_atom());
+  if ( value.is_null() )
+    return false;
+  return A2.unify_term(value);
+}
+
+PREDICATE(atom_term_insert, 2)
+{ map_atom_term.insert(A1.as_atom(), A2);
+  return true;
+}
+
+PREDICATE(atom_term_erase, 1)
+{ map_atom_term.erase(A1.as_atom());
+  return true;
+}
+
+PREDICATE(atom_term_size, 1)
+{ return A1.unify_integer(map_atom_term.size());
+}
