@@ -41,6 +41,9 @@
 #include <SWI-Prolog.h>
 #include <SWI-Stream.h>
 #include <Python.h>
+#if O_PYFRAME
+#include <frameobject.h>	/* Limited API */
+#endif
 #include <assert.h>
 
 static atom_t ATOM_none;
@@ -55,6 +58,8 @@ static atom_t ATOM_file;
 static atom_t ATOM_eval;
 static atom_t ATOM_single;
 static atom_t ATOM_builtins;
+static atom_t ATOM_locals;
+static atom_t ATOM_globals;
 
 static functor_t FUNCTOR_python_error3;
 static functor_t FUNCTOR_error2;
@@ -306,15 +311,15 @@ write_python_object(IOSTREAM *s, atom_t symbol, int flags)
 	name = PyUnicode_AsUTF8AndSize(cname, NULL);
       else
 	name = "noclass";
+
+      SfprintfX(s, "<py_%Us>(%p)", name, obj);
+
+      Py_CLEAR(cls);
+      Py_CLEAR(cname);
       py_gil_release(state);
     } else
-    { name = "no-GIL";
+    { SfprintfX(s, "<py_obj(no-GIL)>(%p)", obj);
     }
-
-    SfprintfX(s, "<py_%Us>(%p)", name, obj);
-
-    Py_CLEAR(cls);
-    Py_CLEAR(cname);
   } else
   { Sfprintf(s, "<py_FREED>(0x0)");
   }
@@ -633,7 +638,7 @@ py_unify_sequence(term_t t, PyObject *obj, int flags)
   term_t head = PL_new_term_ref();
 
   for(ssize_t i=0; i<len; i++)
-  { PyObject *el = check_error(PySequence_GetItem(obj, i));
+  { PyObject *el = PySequence_GetItem(obj, i);
 
     if ( !el )
       return FALSE;
@@ -646,7 +651,6 @@ py_unify_sequence(term_t t, PyObject *obj, int flags)
   if ( !PL_unify_nil(tail) )
     return FALSE;
 
-  PL_reset_term_refs(tail);
   return TRUE;
 }
 
@@ -656,7 +660,7 @@ py_unify_iter(term_t t, PyObject *obj, int flags)
   term_t head = PL_new_term_ref();
   PyObject *item;
 
-  while((item=check_error(PyIter_Next(obj))))
+  while((item=PyIter_Next(obj)))
   { int rc = ( PL_unify_list(tail, head,tail) &&
 	       py_unify(head, item, flags) );
     Py_DECREF(item);
@@ -666,13 +670,12 @@ py_unify_iter(term_t t, PyObject *obj, int flags)
   if ( PL_exception(0) || !PL_unify_nil(tail) )
     return FALSE;
 
-  PL_reset_term_refs(tail);
   return TRUE;
 }
 
 static int
 py_unify_set(term_t t, PyObject *obj, int flags)
-{ PyObject *iter = check_error(PyObject_GetIter(obj));
+{ PyObject *iter = PyObject_GetIter(obj);
   PyObject *item = NULL;
   int rc = FALSE;
 
@@ -745,8 +748,7 @@ py_unify_dict(term_t t, PyObject *obj, int flags)
 	goto out;
       }
     } else
-    { PL_representation_error("py_dict_key");
-      goto out;
+    { goto out;
     }
     if ( !py_unify(pl_values+pli, py_value, flags) )
       goto out;
@@ -772,7 +774,10 @@ py_unify_constant(term_t t, atom_t c)
 
 static int
 py_unify(term_t t, PyObject *obj, int flags)
-{ if ( !obj )
+{ int rc = -1;
+  fid_t fid;
+
+  if ( !obj )
   { check_error(obj);
     return FALSE;
   }
@@ -782,43 +787,58 @@ py_unify(term_t t, PyObject *obj, int flags)
   if ( PyBool_Check(obj) )
     return py_unify_constant(t, PyLong_AsLong(obj) ? ATOM_true : ATOM_false);
 
-  if ( (flags&PYU_OBJ) )	/* py_object(true) effective */
-  { if ( PyLong_CheckExact(obj) )
-      return py_unify_long(t, obj);
-    if ( PyFloat_CheckExact(obj) )
-      return PL_unify_float(t, PyFloat_AsDouble(obj));
-    if ( PyUnicode_CheckExact(obj) )
-      return py_unify_unicode(t, obj, flags);
-    if ( PyTuple_CheckExact(obj) )
-      return py_unify_tuple(t, obj, flags);
-  } else
-  { if ( PyLong_Check(obj) )
-      return py_unify_long(t, obj);
-    if ( PyFloat_Check(obj) )
-      return PL_unify_float(t, PyFloat_AsDouble(obj));
-    if ( PyUnicode_Check(obj) )
-      return py_unify_unicode(t, obj, flags);
-    if ( PyTuple_Check(obj) )
-      return py_unify_tuple(t, obj, flags);
-    if ( PyDict_Check(obj) )
-      return py_unify_dict(t, obj, flags);
-    if ( PyIter_Check(obj) )
-      return py_unify_iter(t, obj, flags);
-    if ( PySequence_Check(obj) )
-      return py_unify_sequence(t, obj, flags);
-    if ( PySet_Check(obj) )
-      return py_unify_set(t, obj, flags);
-    if ( !(flags&PYU_ERROR) )	/* cannot be called with error around */
-    { if ( PyObject_IsInstance(obj, func_Fraction()) )
-	return py_unify_fraction(t, obj);
-      if ( PyEnum_Check(obj) )
-	return py_unify_enum(t, obj);
-      if ( py_is_record(obj) )
-	return py_unify_record(t, obj);
+  if ( (fid=PL_open_foreign_frame()) )
+  { if ( (flags&PYU_OBJ) )	/* py_object(true) effective */
+    { if ( PyLong_CheckExact(obj) )
+	rc = py_unify_long(t, obj);
+      else if ( PyFloat_CheckExact(obj) )
+	rc = PL_unify_float(t, PyFloat_AsDouble(obj));
+      else if ( PyUnicode_CheckExact(obj) )
+	rc = py_unify_unicode(t, obj, flags);
+      else if ( PyTuple_CheckExact(obj) )
+	rc = py_unify_tuple(t, obj, flags);
+    } else
+    { if ( PyLong_Check(obj) )
+	rc = py_unify_long(t, obj);
+      else if ( PyFloat_Check(obj) )
+	rc = PL_unify_float(t, PyFloat_AsDouble(obj));
+      else if ( PyUnicode_Check(obj) )
+	rc = py_unify_unicode(t, obj, flags);
+      else if ( PyTuple_Check(obj) )
+	rc = py_unify_tuple(t, obj, flags);
+      else if ( PyDict_Check(obj) )
+	rc = py_unify_dict(t, obj, flags);
+      else if ( PyIter_Check(obj) )
+	rc = py_unify_iter(t, obj, flags);
+      else if ( PySequence_Check(obj) )
+	rc = py_unify_sequence(t, obj, flags);
+      else if ( PySet_Check(obj) )
+	rc = py_unify_set(t, obj, flags);
+      else if ( !(flags&PYU_ERROR) )	/* cannot be called with error around */
+      { if ( PyObject_IsInstance(obj, func_Fraction()) )
+	  rc = py_unify_fraction(t, obj);
+	else if ( PyEnum_Check(obj) )
+	  rc = py_unify_enum(t, obj);
+	else if ( py_is_record(obj) )
+	  rc = py_unify_record(t, obj);
+      }
     }
-  }
 
-  return unify_py_obj(t, obj);
+    /* conversion failed on Python error: retry as reference */
+    if ( rc == FALSE && !PL_exception(0) )
+    { PyErr_Clear();
+      PL_rewind_foreign_frame(fid);
+      rc = -1;
+    }
+
+    PL_close_foreign_frame(fid);
+
+    if ( rc == -1 )
+      return unify_py_obj(t, obj);
+
+    return rc;
+  } else
+    return FALSE;
 }
 
 
@@ -1497,6 +1517,67 @@ builtin_function(PyObject *builtins, atom_t fname)
 }
 
 
+/* Get the locals()  or globals().  This is pretty much  useless as is
+ * as we  get the  locals and  globals of the  running method  that is
+ * defined in janus.py.
+ *
+ * With CPython  3.11 we can  walk up the  contexts and poke  into our
+ * calling  context.   Using  `O_PYFRAME`  we  get  very  partial  and
+ * untested code trying  to do this.  Notably, we must  find the frame
+ * that is _calling_ the Janus functions.
+ */
+
+static PyObject *
+py_scope(atom_t fname)
+{ PyObject *py_res;
+
+#if O_PYFRAME
+  PyFrameObject *frame = PyEval_GetFrame();
+  int up = 1;
+
+  while(up>0 && frame)
+  { PyFrameObject *fr2 = PyFrame_GetBack(frame);
+    if ( fr2 )
+    { frame = fr2;
+      up--;
+    } else
+      break;
+  }
+
+  if ( fname == ATOM_globals )
+  { py_res = PyFrame_GetGlobals(frame);
+
+    if ( !py_res )		/* no frame */
+    { PyObject *m;
+
+      if ( !(m=check_error(PyImport_AddModule("__main__"))) )
+	return NULL;
+      py_res = PyModule_GetDict(m);
+    }
+  } else
+  { py_res = check_error(PyFrame_GetLocals(frame));
+  }
+#else
+  if ( fname == ATOM_globals )
+  { py_res = PyEval_GetGlobals();
+
+    if ( !py_res )		/* no frame */
+    { PyObject *m;
+
+      if ( !(m=check_error(PyImport_AddModule("__main__"))) )
+	return NULL;
+      py_res = PyModule_GetDict(m);
+    }
+  } else
+  { py_res = check_error(PyEval_GetLocals());
+  }
+#endif
+  if ( py_res )
+    Py_INCREF(py_res);
+
+  return py_res;
+}
+
 /* Evaluate func on obj.  If obj = NULL, evaluate a builtin function
  * Return value: New Reference
  */
@@ -1518,16 +1599,17 @@ py_eval(PyObject *obj, term_t func)
     PyObject *py_argv = NULL;
     PyObject *py_kws  = NULL;
 
-    if ( obj == PyEval_GetBuiltins() )
-    { py_func = builtin_function(obj, fname);
-    } else if ( obj )
+    if ( obj && obj != PyEval_GetBuiltins() )
     { PL_STRINGS_MARK();
       char *fn;
       if ( PL_atom_mbchars(fname, NULL, &fn, REP_UTF8|CVT_EXCEPTION) )
 	py_func = check_error(PyObject_GetAttrString(obj, fn));
       PL_STRINGS_RELEASE();
     } else
-    { py_func = builtin_function(NULL, fname);
+    { if ( fname == ATOM_globals || fname == ATOM_locals )
+	return py_scope(fname);		/* locals() or globals() */
+
+      py_func = builtin_function(obj, fname);
     }
     if ( !py_func )
       goto out;
@@ -1620,18 +1702,13 @@ unchain(term_t call, PyObject **py_target)
   { _PL_get_arg(1, call, on);
     _PL_get_arg(2, call, call);
 
-    if ( !*py_target )
-    { if ( !(rc=get_py_initial_target(on, py_target, TRUE)) )
-	break;
-    } else
-    { PyObject *next = py_eval(*py_target, on);
+    PyObject *next = py_eval(*py_target, on);
 
-      Py_XDECREF(*py_target);
-      *py_target = next;
-      if ( !next )
-      { rc = FALSE;
-	break;
-      }
+    Py_XDECREF(*py_target);
+    *py_target = next;
+    if ( !next )
+    { rc = FALSE;
+      break;
     }
   }
 
@@ -2162,6 +2239,8 @@ install_janus(void)
   MKATOM(eval);
   MKATOM(single);
   MKATOM(builtins);
+  MKATOM(locals);
+  MKATOM(globals);
   ATOM_tuple  = PL_new_atom("-");
   ATOM_pydict = PL_new_atom("py");
   ATOM_curl   = PL_new_atom("{}");
