@@ -60,7 +60,7 @@ extern "C" {
 /* PLVERSION_TAG: a string, normally "", but for example "rc1" */
 
 #ifndef PLVERSION
-#define PLVERSION 90114
+#define PLVERSION 90118
 #endif
 #ifndef PLVERSION_TAG
 #define PLVERSION_TAG ""
@@ -306,6 +306,7 @@ typedef union
 		*    DETERMINISTIC CALL/RETURN  *
 		*********************************/
 
+/* PL_succeed and PL_fail are deprecated */
 #define	PL_succeed	return TRUE	/* succeed deterministically */
 #define PL_fail		return FALSE	/* fail */
 
@@ -919,6 +920,7 @@ PL_EXPORT(int)		_PL_get_arg(int index, term_t t, term_t a);
 #define CVT_WRITE_CANONICAL 0x00000100
 #define CVT_WRITEQ	    0x00000200
 #define CVT_ALL		    (CVT_ATOMIC|CVT_LIST)
+#define CVT_XINTEGER	   (0x00000400|CVT_INTEGER)
 #define CVT_MASK	    0x00000fff
 
 #define CVT_EXCEPTION	    0x00001000	/* throw exception on error */
@@ -1377,11 +1379,6 @@ PL_EXPORT(void)		PL_prof_exit(void *node);
 		 *	      DEBUG		*
 		 *******************************/
 
-PL_EXPORT_DATA(int)     plugin_is_GPL_compatible;
-#ifndef EMACS_MODULE_H
-PL_EXPORT(int)          emacs_module_init(void*);
-#endif
-
 PL_EXPORT(int)		PL_prolog_debug(const char *topic);
 PL_EXPORT(int)		PL_prolog_nodebug(const char *topic);
 
