@@ -3,9 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2011-2020, University of Amsterdam
+    Copyright (c)  2011-2023, University of Amsterdam
                               VU University Amsterdam
 			      CWI, Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -296,7 +297,7 @@ compile_pattern(compiled_pattern *Out, const char *p, int curl, int mflags)
 
 
 static int
-matchPattern(char *s, compiled_pattern *cbuf, int flags)
+matchPattern(const char *s, compiled_pattern *cbuf, int flags)
 { return match_pattern(baseBuffer(&cbuf->pattern, matchcode), s, flags);
 }
 
@@ -556,7 +557,8 @@ utf8_exists_file(DECL_LD const char *name)
 
   return rc;
 #else
-  return AccessFile(name, ACCESS_EXIST);
+  // _xos_access_dir simply uses _waccess().  That is good enough for us.
+  return _xos_access_dir(name, F_OK) == 0 ? TRUE : FALSE;
 #endif
 }
 
@@ -701,11 +703,7 @@ expand(const char *pattern, GlobInfo info)
 	  path[plen++] = '/';
 
 	for(e=readdir(d); e; e = readdir(d))
-	{
-#ifdef __MSDOS__
-	  strlwr(e->d_name);
-#endif
-	  if ( (dot || e->d_name[0] != '.') &&
+	{ if ( (dot || e->d_name[0] != '.') &&
 	       matchPattern(e->d_name, &info->pattern, mflags) )
 	  { char newp[PATH_MAX];
 
@@ -719,6 +717,8 @@ expand(const char *pattern, GlobInfo info)
 #endif
 	    }
 	  }
+	  if ( PL_handle_signals() < 0 )
+	    return FALSE;
 	}
 	closedir(d);
       }
@@ -731,6 +731,23 @@ expand(const char *pattern, GlobInfo info)
 }
 
 
+static int
+utf8casecmp(const char *s1, const char *s2)
+{ while(*s1 && *s2)
+  { int c1, c2;
+
+    s1 = utf8_get_char(s1, &c1);
+    s2 = utf8_get_char(s2, &c2);
+    c1 = makeLowerW(c1);
+    c2 = makeLowerW(c2);
+
+    if ( c1 != c2 )
+      return c1-c2;
+  }
+
+  return *s1 && *s2 ? 0 : *s1 ? 1 : -1;
+}
+
 static int
 compareBagEntries(const void *a1, const void *a2)
 { GET_LD
@@ -743,9 +760,9 @@ compareBagEntries(const void *a1, const void *a2)
   s2 = expand_str(info, i2);
 
   if ( truePrologFlag(PLFLAG_FILE_CASE) )
-    return mbscoll(s1, s2);
+    return strcmp(s1, s2);
   else
-    return mbscasecoll(s1, s2);
+    return utf8casecmp(s1, s2);
 }
 
 
@@ -834,7 +851,7 @@ PRED_IMPL("directory_files", 2, directory_files, 0)
     return PL_unify_nil(tail);
   }
 
-  return PL_error(NULL, 0, OsError(), ERR_FILE_OPERATION,
+  return PL_error(NULL, 0, MSG_ERRNO, ERR_FILE_OPERATION,
 		  ATOM_open, ATOM_directory, A1);
 }
 
