@@ -3,9 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2005-2020, University of Amsterdam
-                              VU University Amsterdam
+    Copyright (c)  2005-2023, University of Amsterdam
+			      VU University Amsterdam
 			      CWI, Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -39,37 +40,65 @@
 #ifndef O_PLGMP_INCLUDED
 #define O_PLGMP_INCLUDED
 
-int	PL_unify_number__LD(term_t t, Number n ARG_LD);
-int	PL_put_number__LD(term_t t, Number n ARG_LD);
-void	get_number(word w, Number n  ARG_LD);
-int	PL_get_number__LD(term_t t, Number n ARG_LD);
+#if USE_LD_MACROS
+#define	PL_unify_number(t, n)		LDFUNC(PL_unify_number, t, n)
+#define	PL_put_number(t, n)		LDFUNC(PL_put_number, t, n)
+#define	get_number(w, n)		LDFUNC(get_number, w, n)
+#define	PL_get_number(t, n)		LDFUNC(PL_get_number, t, n)
+#define	put_uint64(at, l, flags)	LDFUNC(put_uint64, at, l, flags)
+#define	put_number(at, n, flags)	LDFUNC(put_number, at, n, flags)
+#ifdef O_BIGNUM
+#define	get_rational_no_int(w, n)	LDFUNC(get_rational_no_int, w, n)
+#endif
+#endif /*USE_LD_MACROS*/
+
+#define LDFUNC_DECLARATIONS
+
+int	PL_unify_number(term_t t, Number n);
+int	PL_put_number(term_t t, Number n);
+void	get_number(word w, Number n);
 int	PL_get_number(term_t t, Number n);
-int	put_uint64(Word at, uint64_t l, int flags ARG_LD);
-int	put_number(Word at, Number n, int flags ARG_LD);
+int	PL_get_number(term_t t, Number n);
+int	put_uint64(Word at, uint64_t l, int flags);
+int	put_number(Word at, Number n, int flags);
 int	promoteToFloatNumber(Number n);
 int	make_same_type_numbers(Number n1, Number n2) WUNUSED;
 int     promoteNumber(Number n1, numtype type) WUNUSED;
 int	cmpNumbers(Number n1, Number n2);
+int	cmpReals(Number n1, Number n2);
 void	cpNumber(Number to, Number from);
+#ifdef O_BIGNUM
+void	get_rational_no_int(word w, number *n);
+#endif
 
-		 /*******************************
-		 *	LD-USING FUNCTIONS	*
-		 *******************************/
+#undef LDFUNC_DECLARATIONS
 
-#define PL_get_number(t, n)	PL_get_number__LD(t, n PASS_LD)
-#define PL_unify_number(t, n)	PL_unify_number__LD(t, n PASS_LD)
-#define PL_put_number(t, n)	PL_put_number__LD(t, n PASS_LD)
-
-#ifdef O_GMP
+#if O_BIGNUM
+#if O_GMP
 #include <gmp.h>
 
+#define MPZ_ON_STACK(n)		(!((n)->_mp_alloc))
+#define MPZ_SET_READONLY(n)	((n)->_mp_alloc = 0)
+#define MPZ_LIMB_SIZE(n)	((n)->_mp_size)
+#define MPZ_LIMBS(n)		((n)->_mp_d)
+#define MPZ_STACK_EXTRA		(1)
+#elif O_BF
+#include "libbf/bf_gmp.h"
+#include "pl-bf.h"
+
+#define MPZ_ON_STACK(n)		(!(n->ctx))
+#define MPZ_SET_READONLY(n)	((n)->ctx = NULL)
+#define MPZ_LIMB_SIZE(n)	((n)->sign ? -(n)->len : (n)->len)
+#define MPZ_LIMBS(n)		((n)->tab)
+#define MPZ_STACK_EXTRA		(2)
+#endif
+
 #define O_MY_GMP_ALLOC 1
 #define O_GMP_PRECHECK_ALLOCATIONS 1	/* GMP 4.2.3 uses abort() sometimes */
 
 void	initGMP(void);
 void	cleanupGMP(void);
 void	get_integer(word w, number *n);
-void	get_rational(word w, number *n);
 Code	get_mpz_from_code(Code pc, mpz_t mpz);
 Code	get_mpq_from_code(Code pc, mpq_t mpq);
 int	promoteToMPZNumber(number *n);
@@ -85,12 +114,21 @@ char *	skipMPQOnCharp(const char *data);
 int	mpz_to_int64(mpz_t mpz, int64_t *i);
 int	mpz_to_uint64(mpz_t mpz, uint64_t *i);
 void	mpz_init_set_si64(mpz_t mpz, int64_t i);
-double	mpX_round(double f);
+double	mpz_to_double(mpz_t n);
 double	mpq_to_double(mpq_t q);
 void	mpq_set_double(mpq_t q, double f);
+word	bignum_index(const word *p);
 
-#define clearNumber(n) \
-	do { if ( (n)->type != V_INTEGER ) clearGMPNumber(n); } while(0)
+static inline void
+clearNumber(Number n)
+{ switch(n->type)
+  { case V_INTEGER:
+    case V_FLOAT:
+      return;
+    default:
+      clearGMPNumber(n);
+  }
+}
 
 static inline word
 mpz_size_stack(int sz)
@@ -112,7 +150,28 @@ mpq_stack_size(word w)
 { return (int)w>>1;
 }
 
-#else /*O_GMP*/
+#ifdef O_GMP
+static inline void
+mpz_add_si(mpz_t r, const mpz_t n1, long add)
+{ if ( add > 0 )
+    mpz_add_ui(r, n1, add);
+  else
+    mpz_sub_ui(r, n1, -add);
+}
+#endif
+
+#define get_rational(w, n) LDFUNC(get_rational, w, n)
+static inline void
+get_rational(DECL_LD word w, Number n)
+{ if ( storage(w) == STG_INLINE )
+  { n->value.i = valInt(w);
+    n->type = V_INTEGER;
+  } else
+  { get_rational_no_int(w, n);
+  }
+}
+
+#else /*O_BIGNUM*/
 
 #define get_integer(w, n) \
 	do \
@@ -127,7 +186,7 @@ mpq_stack_size(word w)
 #define ensureWritableNumber(n) (void)0
 #define initGMP()		(void)0
 
-#endif /*O_GMP*/
+#endif /*O_BIGNUM*/
 
 
 		 /*******************************
