@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2021, University of Amsterdam
+    Copyright (c)  1985-2023, University of Amsterdam
 			      VU University Amsterdam
 			      CWI, Amsterdam
 			      SWI-Prolog Solutions b.v.
@@ -54,6 +54,10 @@
 #include "pl-proc.h"
 #include "os/pl-prologflag.h"
 #include <stdio.h>
+#ifdef __WINDOWS__
+#include "pl-nt.h"
+#include <process.h>
+#endif
 
 #define WFG_TRACING	0x02000
 #define WFG_BACKTRACE	0x04000
@@ -230,9 +234,9 @@ exitFromDebugger(const char *msg, int status)
 { GET_LD
 
 #ifdef O_PLMT
-  if ( PL_thread_self() > 1 && !LD->exit_requested )
+  if ( PL_thread_self() > 1 && !LD->thread.exit_requested )
   { Sfprintf(Sdout, "%sexit session\n", msg);
-    LD->exit_requested = EXIT_REQ_THREAD;
+    LD->thread.exit_requested = EXIT_REQ_THREAD;
     return ACTION_ABORT;
   }
 #endif
@@ -513,7 +517,7 @@ again:
       buf[0] = c;
       buf[1] = EOS;
       if ( isDigit(buf[0]) || buf[0] == '/' || buf[0] == '-' )
-      { Sfprintf(Sdout, buf);
+      { Sfprintf(Sdout, "%s", buf);
 	readLine(Sdin, Sdout, buf);
       }
     }
@@ -1022,7 +1026,7 @@ writeFrameGoal(IOSTREAM *out, LocalFrame frame, Code PC, unsigned int flags)
     ctx.context = 0;
     ctx.control = FRG_FIRST_CALL;
     ctx.engine  = LD;
-    if ( !pl_prolog_flag(tmp, options, &ctx) )
+    if ( !pl_prolog_flag5(tmp, options, 0, 0, 0, &ctx) )
       PL_put_nil(options);
     PL_unify_stream_or_alias(tmp, out);
 
@@ -1252,7 +1256,7 @@ _PL_backtrace(IOSTREAM *out, int depth, int flags)
       if ( frame->predicate == def )
       { if ( ++same_proc >= 10 )
 	{ if ( same_proc == 10 )
-	    Sfprintf(out, "    ...\n    ...\n", Sdout);
+	    Sfprintf(out, "    ...\n    ...\n");
 	  rctx = ctx;
 	  continue;
 	}
@@ -1454,8 +1458,8 @@ traceInterception(LocalFrame frame, Choice bfr, int port, Code PC)
     }
 
   out:
-    if ( qid ) PL_close_query(qid);
-    if ( cid ) PL_discard_foreign_frame(cid);
+    if ( qid ) PL_cut_query(qid);
+    if ( cid ) PL_close_foreign_frame(cid);
 
     if ( nodebug )
     { tracemode(FALSE, NULL);
@@ -1563,7 +1567,7 @@ PL_describe_context(pl_context_t *c, char *buf, size_t len)
     int printed;
 
     if ( !onStack(local, fr) )
-      return snprintf(buf, len, "<invalid frame reference %p>", fr);
+      return snprintf(buf, len, "<invalid frame reference %p>", (void*)fr);
 
     level = levelFrame(fr);
     if ( !fr->predicate )
@@ -1772,7 +1776,7 @@ interruptHandler(int sig)
     return;
   }
 
-  if ( LD->exit_requested )
+  if ( LD->thread.exit_requested )
   { term_t rval = PL_new_term_ref();
     PL_put_atom(rval, ATOM_true);
     pl_thread_exit(rval);
@@ -2317,6 +2321,7 @@ static int
 prolog_frame_attribute(term_t frame, term_t what, term_t value)
 { GET_LD
   LocalFrame fr;
+  term_t fref;			/* Save/restore fr over possible shifts */
   atom_t key;
   size_t arity;
   term_t result = PL_new_term_ref();
@@ -2326,6 +2331,7 @@ prolog_frame_attribute(term_t frame, term_t what, term_t value)
     return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_frame_reference, frame);
   if ( !fr )
     return FALSE;				/* frame == 'none' */
+  fref = consTermRef((Word)fr);
   if ( !PL_get_name_arity(what, &key, &arity) )
     return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_callable, what);
   if ( !PL_strip_module(value, &m, value) )
@@ -2355,15 +2361,15 @@ prolog_frame_attribute(term_t frame, term_t what, term_t value)
     checkData(argFrameP(fr, argn-1));
 #endif
 
-   if ( !hasGlobalSpace(0) )
-   { int rc;
+    if ( !hasGlobalSpace(0) )
+    { int rc;
 
-     if ( (rc=ensureGlobalSpace(0, ALLOW_GC)) != TRUE )
-       return raiseStackOverflow(rc);
-     PL_get_frame(frame, &fr);
-   }
+      if ( (rc=ensureGlobalSpace(0, ALLOW_GC)) != TRUE )
+	return raiseStackOverflow(rc);
+      fr = (LocalFrame)valTermRef(fref);
+    }
 
-   return PL_unify(value, consTermRef(argFrameP(fr, argn-1)));
+    return PL_unify(value, consTermRef(argFrameP(fr, argn-1)));
   }
 
   if ( !(arity == 0 || (arity == 1 && key == ATOM_parent_goal)) )
@@ -2422,13 +2428,18 @@ prolog_frame_attribute(term_t frame, term_t what, term_t value)
       fail;
 
     if ( (fid = PL_open_foreign_frame()) )
-    { while( fr )
+    { fr = (LocalFrame)valTermRef(fref);
+
+      while( fr )
       { while(fr && fr->predicate != proc->definition)
-	  fr = parentFrame(fr);
+	{ fr = parentFrame(fr);
+	  assert(!fr || isFrame(fr));
+	}
 
 	if ( fr )
 	{ int i, garity = fr->predicate->functor->arity;
-	  term_t fref = consTermRef(fr);
+
+	  fref = consTermRef((Word)fr);
 
 	  for(i=0; i<garity; i++)
 	  { term_t fa;
@@ -2437,15 +2448,17 @@ prolog_frame_attribute(term_t frame, term_t what, term_t value)
 
 	    _PL_get_arg(i+1, head, a);
 	    if ( !PL_unify(a, fa) )
-	      break;
+	      break;				/* Argument does not unify */
 	    fr = (LocalFrame)valTermRef(fref);	/* deal with possible shift */
 	  }
-	  if ( i == garity )
-	  { if ( arity == 1 )
+	  /* fr is valid because restored after every unification */
+	  if ( i == garity )			/* All arguments unified: found a match */
+	  { if ( arity == 1 )			/* parent_frame(Next) */
 	    { LocalFrame parent;
 	      term_t arg = PL_new_term_ref();
 
 	      _PL_get_arg(1, what, arg);
+	      fr = (LocalFrame)valTermRef(fref);
 	      if ( (parent = parentFrame(fr)) )
 	      { if ( PL_unify_frame(arg, parent) )
 		  return TRUE;
@@ -2468,10 +2481,11 @@ prolog_frame_attribute(term_t frame, term_t what, term_t value)
 	  }
 	} else
 	{ PL_close_foreign_frame(fid);
-	  return FALSE;
+	  return FALSE;				/* No frame running this predicate */
 	}
       }
-    }
+    } else
+      return FALSE;
   } else if ( key == ATOM_pc )
   { if ( fr->programPointer &&
 	 fr->parent &&
